@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2008-2013  Internet Systems Consortium, Inc. ("ISC")
+ * Copyright (C) 2008-2015  Internet Systems Consortium, Inc. ("ISC")
  *
  * Permission to use, copy, modify, and/or distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -14,8 +14,6 @@
  * PERFORMANCE OF THIS SOFTWARE.
  */
 
-/* $Id: statschannel.c,v 1.28 2011/03/12 04:59:46 tbox Exp $ */
-
 /*! \file */
 
 #include <config.h>
@@ -27,6 +25,7 @@
 #include <isc/print.h>
 #include <isc/socket.h>
 #include <isc/stats.h>
+#include <isc/string.h>
 #include <isc/task.h>
 
 #include <dns/cache.h>
@@ -122,6 +121,8 @@ set_desc(int counter, int maxcounter, const char *fdesc, const char **fdescs,
 	 const char *xdesc, const char **xdescs)
 {
 	REQUIRE(counter < maxcounter);
+if (fdescs[counter] != NULL)
+fprintf(stderr, "fdescs[%d] == %s\n", counter, fdescs[counter]);
 	REQUIRE(fdescs[counter] == NULL);
 #ifdef HAVE_LIBXML2
 	REQUIRE(xdescs[counter] == NULL);
@@ -167,7 +168,7 @@ init_desc(void) {
 	SET_NSSTATDESC(sig0in, "requests with SIG(0) received", "ReqSIG0");
 	SET_NSSTATDESC(invalidsig, "requests with invalid signature",
 		       "ReqBadSIG");
-	SET_NSSTATDESC(tcp, "TCP requests received", "ReqTCP");
+	SET_NSSTATDESC(requesttcp, "TCP requests received", "ReqTCP");
 	SET_NSSTATDESC(authrej, "auth queries rejected", "AuthQryRej");
 	SET_NSSTATDESC(recurserej, "recursive queries rejected", "RecQryRej");
 	SET_NSSTATDESC(xfrrej, "transfer requests rejected", "XfrRej");
@@ -206,8 +207,32 @@ init_desc(void) {
 	SET_NSSTATDESC(updatebadprereq,
 		       "updates rejected due to prerequisite failure",
 		       "UpdateBadPrereq");
+	SET_NSSTATDESC(recursclients, "recursing clients",
+			"RecursClients");
+	SET_NSSTATDESC(dns64, "queries answered by DNS64", "DNS64");
+	SET_NSSTATDESC(ratedropped, "responses dropped for rate limits",
+		       "RateDropped");
+	SET_NSSTATDESC(rateslipped, "responses truncated for rate limits",
+		       "RateSlipped");
 	SET_NSSTATDESC(rpz_rewrites, "response policy zone rewrites",
 		       "RPZRewrites");
+	SET_NSSTATDESC(udp, "UDP queries received", "QryUDP");
+	SET_NSSTATDESC(tcp, "TCP queries received", "QryTCP");
+	SET_NSSTATDESC(nsidopt, "NSID option received", "NSIDOpt");
+	SET_NSSTATDESC(expireopt, "Expire option received", "ExpireOpt");
+	SET_NSSTATDESC(otheropt, "Other EDNS option received", "OtherOpt");
+	SET_NSSTATDESC(sitopt, "source identity token option received",
+		       "SitOpt");
+	SET_NSSTATDESC(sitnew, "new source identity token requested",
+		       "SitNew");
+	SET_NSSTATDESC(sitbadsize, "source identity token - bad size",
+		       "SitBadSize");
+	SET_NSSTATDESC(sitbadtime, "source identity token - bad time",
+		       "SitBadTime");
+	SET_NSSTATDESC(sitnomatch, "source identity token - no match",
+		       "SitNoMatch");
+	SET_NSSTATDESC(sitmatch, "source identity token - match", "SitMatch");
+	SET_NSSTATDESC(ecsopt, "EDNS client subnet option recieved", "ECSOpt");
 	INSIST(i == dns_nsstatscounter_max);
 
 	/* Initialize resolver statistics */
@@ -278,6 +303,10 @@ init_desc(void) {
 	SET_RESSTATDESC(queryrtt5, "queries with RTT > "
 			DNS_RESOLVER_QRYRTTCLASS4STR "ms",
 			"QryRTT" DNS_RESOLVER_QRYRTTCLASS4STR "+");
+	SET_RESSTATDESC(zonequota, "spilled due to zone quota", "ZoneQuota");
+	SET_RESSTATDESC(serverquota, "spilled due to server quota",
+			"ServerQuota");
+
 	INSIST(i == dns_resstatscounter_max);
 
 	/* Initialize zone statistics */
@@ -488,7 +517,7 @@ dump_counters(isc_stats_t *stats, statsformat_t type, void *arg,
 	      const char *category, const char **desc, int ncounters,
 	      int *indices, isc_uint64_t *values, int options)
 {
-	int i, index;
+	int i, idx;
 	isc_uint64_t value;
 	stats_dumparg_t dumparg;
 	FILE *fp;
@@ -510,8 +539,8 @@ dump_counters(isc_stats_t *stats, statsformat_t type, void *arg,
 	isc_stats_dump(stats, generalstat_dump, &dumparg, options);
 
 	for (i = 0; i < ncounters; i++) {
-		index = indices[i];
-		value = values[index];
+		idx = indices[i];
+		value = values[idx];
 
 		if (value == 0 && (options & ISC_STATSDUMP_VERBOSE) == 0)
 			continue;
@@ -520,7 +549,7 @@ dump_counters(isc_stats_t *stats, statsformat_t type, void *arg,
 		case statsformat_file:
 			fp = arg;
 			fprintf(fp, "%20" ISC_PRINT_QUADFORMAT "u %s\n",
-				value, desc[index]);
+				value, desc[idx]);
 			break;
 		case statsformat_xml:
 #ifdef HAVE_LIBXML2
@@ -538,7 +567,7 @@ dump_counters(isc_stats_t *stats, statsformat_t type, void *arg,
 						       "name"));
 			TRY0(xmlTextWriterWriteString(writer,
 						      ISC_XMLCHAR
-						      desc[index]));
+						      desc[i]));
 			TRY0(xmlTextWriterEndElement(writer));
 			/* </name> */
 
@@ -562,7 +591,7 @@ dump_counters(isc_stats_t *stats, statsformat_t type, void *arg,
 							 ISC_XMLCHAR
 							 "name",
 							 ISC_XMLCHAR
-							 desc[index]));
+							 desc[i]));
 			TRY0(xmlTextWriterWriteFormatString(writer,
 				"%" ISC_PRINT_QUADFORMAT "u", value));
 			TRY0(xmlTextWriterEndElement(writer));
@@ -580,7 +609,7 @@ dump_counters(isc_stats_t *stats, statsformat_t type, void *arg,
 							       "name"));
 				TRY0(xmlTextWriterWriteString(writer,
 							      ISC_XMLCHAR
-							      desc[index]));
+							      desc[idx]));
 				TRY0(xmlTextWriterEndElement(writer)); /* name */
 
 				TRY0(xmlTextWriterStartElement(writer,
@@ -589,7 +618,7 @@ dump_counters(isc_stats_t *stats, statsformat_t type, void *arg,
 			} else {
 				TRY0(xmlTextWriterStartElement(writer,
 							       ISC_XMLCHAR
-							       desc[index]));
+							       desc[idx]));
 			}
 			TRY0(xmlTextWriterWriteFormatString(writer,
 							    "%"
@@ -886,7 +915,6 @@ static isc_result_t
 zone_xmlrender(dns_zone_t *zone, void *arg) {
 	isc_result_t result;
 	char buf[1024 + 32];	/* sufficiently large for zone name and class */
-	char *zone_name_only = NULL;
 	dns_rdataclass_t rdclass;
 	isc_uint32_t serial;
 	xmlTextWriterPtr writer = arg;
@@ -905,13 +933,11 @@ zone_xmlrender(dns_zone_t *zone, void *arg) {
 	dumparg.arg = writer;
 
 	TRY0(xmlTextWriterStartElement(writer, ISC_XMLCHAR "zone"));
-	dns_zone_name(zone, buf, sizeof(buf));
-	zone_name_only = strtok(buf, "/");
-	if(zone_name_only == NULL)
-		zone_name_only = buf;
 
+	dns_zone_nameonly(zone, buf, sizeof(buf));
 	TRY0(xmlTextWriterWriteAttribute(writer, ISC_XMLCHAR "name",
-					 ISC_XMLCHAR zone_name_only));
+					 ISC_XMLCHAR buf));
+
 	rdclass = dns_zone_getclass(zone);
 	dns_rdataclass_format(rdclass, buf, sizeof(buf));
 	TRY0(xmlTextWriterWriteAttribute(writer, ISC_XMLCHAR "rdataclass",
@@ -972,13 +998,18 @@ zone_xmlrender(dns_zone_t *zone, void *arg) {
 	isc_uint32_t serial;
 	xmlTextWriterPtr writer = arg;
 	isc_stats_t *zonestats;
+	dns_zonestat_level_t statlevel;
 	isc_uint64_t nsstat_values[dns_nsstatscounter_max];
 	int xmlrc;
 	isc_result_t result;
 
+	statlevel = dns_zone_getstatlevel(zone);
+	if (statlevel == dns_zonestat_none)
+		return (ISC_R_SUCCESS);
+
 	TRY0(xmlTextWriterStartElement(writer, ISC_XMLCHAR "zone"));
 
-	dns_zone_name(zone, buf, sizeof(buf));
+	dns_zone_nameonly(zone, buf, sizeof(buf));
 	TRY0(xmlTextWriterStartElement(writer, ISC_XMLCHAR "name"));
 	TRY0(xmlTextWriterWriteString(writer, ISC_XMLCHAR buf));
 	TRY0(xmlTextWriterEndElement(writer));
@@ -1046,7 +1077,7 @@ generatexml(ns_server_t *server, int *buflen, xmlChar **buf) {
 			ISC_XMLCHAR "type=\"text/xsl\" href=\"/bind9.ver3.xsl\""));
 	TRY0(xmlTextWriterStartElement(writer, ISC_XMLCHAR "statistics"));
 	TRY0(xmlTextWriterWriteAttribute(writer, ISC_XMLCHAR "version",
-					 ISC_XMLCHAR "3.0"));
+					 ISC_XMLCHAR "3.6"));
 
 	/* Set common fields for statistics dump */
 	dumparg.type = statsformat_xml;
@@ -1082,9 +1113,9 @@ generatexml(ns_server_t *server, int *buflen, xmlChar **buf) {
 			if (dumparg.result != ISC_R_SUCCESS)
 				goto error;
 		}
+else fprintf(stderr, "WTF WHERE'S RESQUERYRSTATS\n");
 		TRY0(xmlTextWriterEndElement(writer));
 
-		/* <resstats> */
 		TRY0(xmlTextWriterStartElement(writer, ISC_XMLCHAR "counters"));
 		TRY0(xmlTextWriterWriteAttribute(writer, ISC_XMLCHAR "type",
 						 ISC_XMLCHAR "resstats"));
@@ -1098,7 +1129,7 @@ generatexml(ns_server_t *server, int *buflen, xmlChar **buf) {
 			if (result != ISC_R_SUCCESS)
 				goto error;
 		}
-		TRY0(xmlTextWriterEndElement(writer)); /* </resstats> */
+		TRY0(xmlTextWriterEndElement(writer));
 
 		cacherrstats = dns_db_getrrsetstats(view->cachedb);
 		if (cacherrstats != NULL) {
@@ -1145,7 +1176,7 @@ generatexml(ns_server_t *server, int *buflen, xmlChar **buf) {
 					 ISC_XMLCHAR "opcode"));
 
 	dns_opcodestats_dump(server->opcodestats, opcodestat_dump, &dumparg,
-			     0);
+			     ISC_STATSDUMP_VERBOSE);
 	if (dumparg.result != ISC_R_SUCCESS)
 		goto error;
 
@@ -1232,6 +1263,8 @@ generatexml(ns_server_t *server, int *buflen, xmlChar **buf) {
 	xmlFreeTextWriter(writer);
 
 	xmlDocDumpFormatMemoryEnc(doc, buf, buflen, "UTF-8", 0);
+	if (*buf == NULL)
+		goto error;
 	xmlFreeDoc(doc);
 	return (ISC_R_SUCCESS);
 
@@ -1357,10 +1390,10 @@ generatexml(ns_server_t *server, int *buflen, xmlChar **buf) {
 	TRY0(xmlTextWriterStartElement(writer, ISC_XMLCHAR "server"));
 	TRY0(xmlTextWriterStartElement(writer, ISC_XMLCHAR "boot-time"));
 	TRY0(xmlTextWriterWriteString(writer, ISC_XMLCHAR boottime));
-	TRY0(xmlTextWriterEndElement(writer));
+	TRY0(xmlTextWriterEndElement(writer)); /* boot-time */
 	TRY0(xmlTextWriterStartElement(writer, ISC_XMLCHAR "current-time"));
 	TRY0(xmlTextWriterWriteString(writer, ISC_XMLCHAR nowstr));
-	TRY0(xmlTextWriterEndElement(writer));
+	TRY0(xmlTextWriterEndElement(writer)); /* current-time */
 
 	TRY0(xmlTextWriterStartElement(writer, ISC_XMLCHAR "requests"));
 	dumparg.result = ISC_R_SUCCESS;
@@ -1426,6 +1459,8 @@ generatexml(ns_server_t *server, int *buflen, xmlChar **buf) {
 	xmlFreeTextWriter(writer);
 
 	xmlDocDumpFormatMemoryEnc(doc, buf, buflen, "UTF-8", 1);
+	if (*buf == NULL)
+		goto error;
 	xmlFreeDoc(doc);
 	return (ISC_R_SUCCESS);
 
@@ -1446,18 +1481,21 @@ wrap_xmlfree(isc_buffer_t *buffer, void *arg) {
 }
 
 static isc_result_t
-render_index(const char *url, const char *querystring, void *arg,
+render_index(const char *url, isc_httpdurl_t *urlinfo,
+	     const char *querystring, const char *headers, void *arg,
 	     unsigned int *retcode, const char **retmsg, const char **mimetype,
 	     isc_buffer_t *b, isc_httpdfree_t **freecb,
 	     void **freecb_args)
 {
-	unsigned char *msg;
+	unsigned char *msg = NULL;
 	int msglen;
 	ns_server_t *server = arg;
 	isc_result_t result;
 
 	UNUSED(url);
+	UNUSED(urlinfo);
 	UNUSED(querystring);
+	UNUSED(headers);
 
 	result = generatexml(server, &msglen, &msg);
 
@@ -1480,22 +1518,56 @@ render_index(const char *url, const char *querystring, void *arg,
 #endif	/* HAVE_LIBXML2 */
 
 static isc_result_t
-render_xsl(const char *url, const char *querystring, void *args,
-	   unsigned int *retcode, const char **retmsg, const char **mimetype,
-	   isc_buffer_t *b, isc_httpdfree_t **freecb,
-	   void **freecb_args)
+render_xsl(const char *url, isc_httpdurl_t *urlinfo,
+	   const char *querystring, const char *headers,
+	   void *args, unsigned int *retcode, const char **retmsg,
+	   const char **mimetype, isc_buffer_t *b,
+	   isc_httpdfree_t **freecb, void **freecb_args)
 {
+	isc_result_t result;
+
 	UNUSED(url);
 	UNUSED(querystring);
 	UNUSED(args);
 
-	*retcode = 200;
-	*retmsg = "OK";
-	*mimetype = "text/xslt+xml";
-	isc_buffer_reinit(b, xslmsg, strlen(xslmsg));
-	isc_buffer_add(b, strlen(xslmsg));
 	*freecb = NULL;
 	*freecb_args = NULL;
+	*mimetype = "text/xslt+xml";
+
+	if (urlinfo->isstatic) {
+		isc_time_t when;
+		char *p = strcasestr(headers, "If-Modified-Since: ");
+
+		if (p != NULL) {
+			time_t t1, t2;
+			p += strlen("If-Modified-Since: ");
+			result = isc_time_parsehttptimestamp(p, &when);
+			if (result != ISC_R_SUCCESS)
+				goto send;
+
+			result = isc_time_secondsastimet(&when, &t1);
+			if (result != ISC_R_SUCCESS)
+				goto send;
+
+			result = isc_time_secondsastimet(&urlinfo->loadtime,
+							 &t2);
+			if (result != ISC_R_SUCCESS)
+				goto send;
+
+			if (t1 < t2)
+				goto send;
+
+			*retcode = 304;
+			*retmsg = "Not modified";
+			return (ISC_R_SUCCESS);
+		}
+	}
+
+ send:
+	*retcode = 200;
+	*retmsg = "OK";
+	isc_buffer_reinit(b, xslmsg, strlen(xslmsg));
+	isc_buffer_add(b, strlen(xslmsg));
 
 	return (ISC_R_SUCCESS);
 }
@@ -1632,11 +1704,11 @@ add_listener(ns_server_t *server, ns_statschannel_t **listenerp,
 #endif /* NEWSTATS */
 #endif
 #ifdef NEWSTATS
-	isc_httpdmgr_addurl(listener->httpdmgr, "/bind9.ver3.xsl", render_xsl,
-			    server);
+	isc_httpdmgr_addurl2(listener->httpdmgr, "/bind9.ver3.xsl", ISC_TRUE,
+			     render_xsl, server);
 #else /* OLDSTATS */
-	isc_httpdmgr_addurl(listener->httpdmgr, "/bind9.xsl", render_xsl,
-			    server);
+	isc_httpdmgr_addurl2(listener->httpdmgr, "/bind9.xsl", ISC_TRUE,
+			     render_xsl, server);
 #endif /* NEWSTATS */
 	*listenerp = listener;
 	isc_log_write(ns_g_lctx, NS_LOGCATEGORY_GENERAL,
